import numpy as np
import pandas as pd

def string2nan(df):
    missing_cols = [df.dtypes.index[i] for i, type in enumerate(df.dtypes) if type == "object"]
    for col in missing_cols:
        for i, value in enumerate(df[col]):
            if isfloat(value):
                df[col][i] = np.nan
    return(df)

def isfloat(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def islist(value):
    text = str(value)
    if text[0] == '[' and text[len(text) - 1] == ']':
        return True
    else:
        return False


def istuple(value):
    text = str(value)
    if text[0] == '(' and text[len(text) - 1] == ')':
        return True
    else:
        return False


def isnpdot(value):
    text = str(value)
    if text[0:3] == 'np.':
        return True
    else:
        return False


def isint(val):
    try:
        int(val)
        return True
    except ValueError:
        return False


def value(string):
    try:
        val =  eval(string)
        if type(val)==type:
            return string
        else:
            return val
    except:
        return string

def check_input(X,name,n0=None,n1=None, format_out='df'):
    """
    Makes sure that input is a 2-D numpy array or pandas dataframe in the correct format.

    :param X: numpy.ndarray or pandas.DataFrame
        input data
    :param name: string
        name of input (e.g. training input)
    :param n0: int
        number of data entries
    :param n1: int
        number of features
    :param format_out: string ('df' or 'ar'), optional (default = 'df')

    :return input data converted to array or dataframe
    :return the header of dataframe
        if input data is not a dataframe return None

    """
    if not (X.ndim == 2):
        raise Exception(name+' needs to be two dimensional')
    if isinstance(X, pd.DataFrame):
        if format_out == 'ar':
            if X.shape[1]>1:
                header = X.columns
                X = X.values
            else:
                if n0 == 1:
                    header = X.columns
                    X = X[header[0]].values
                else:
                    header = X.columns
                    X = X.values
        else:
            header = X.columns
        if not np.can_cast(X.dtypes, np.float, casting='same_kind'):
            raise Exception(name + ' cannot be cast to floats')
    elif isinstance(X, np.ndarray):
        if format_out == 'df':
            X = pd.DataFrame(X)
            header = None
        else:
            header = None
    else:
        raise Exception(name+' needs to be either pandas dataframe or numpy array')
    if n0 and X.shape[0] != n0:
        raise Exception(name+' has an invalid number of data entries')
    if n1 and X.shape[1] != n1:
        raise Exception(name+' has an invalid number of feature entries')
    return X.astype(float), header

def check_object_col(df, name):
    """
    Goals:
    - check if columns with type 'object' don't have elements that can be
      converted to numeric values.
    - remove columns with all non numeric elements.
    """
    object_cols = [df.dtypes.index[i] for i, typ in enumerate(df.dtypes) if typ == "object"]
    for col in object_cols:
        for i, value in enumerate(df[col]):
            if  isfloat(value):
                raise ValueError("column '%s' in '%s' includes both string and float values." %(str(col),name))
    # drop object columns
    if len(object_cols)>0:
        df = df.drop(object_cols,1)
    return df



# __name__ = "chemeco"
__version__ = "0.1.0"
__author__ = ["Mojtaba Haghighatlari (mojtabah@buffalo.edu)", "Johannes Hachmann (hachmann@buffalo.edu)"]

from chemeco.wrappers.engine import run as ChemEcoRun
from chemeco.notebooks import ChemEcoNotebook

import sys
sys.dont_write_bytecode = True



from chemeco.wrappers.database import sklearn_db
from chemeco.wrappers.database import cheml_db
from chemeco.wrappers.database import pandas_db
import inspect

def tshf():
    """
    tshf stands for the combination of task, subtask, host, and function
    :return: combination, dictionary of the aforementioned combinations
    """
    # 7 tasks
    tasks = ['Enter', 'Represent', 'Prepare', 'Model', 'Search', 'Mix', 'Visualize', 'Store']
    extras = ['np', '__builtins__', '__doc__', '__file__', '__name__', '__package__', 'mask', 'Input', 'Output',
              'Parameter', 'req', 'regression_types', 'cv_classes']

    combination = {task: {} for task in tasks}
    all_classes = [k[1] for k in inspect.getmembers(sklearn_db) if k[0][0:2]!='__']
    all_classes += [k[1] for k in inspect.getmembers(cheml_db) if k[0][0:2]!='__' ]
    all_classes += [k[1] for k in inspect.getmembers(pandas_db) if k[0][0:2]!='__' ]
    for k in all_classes:
        vk = vars(k)
        if 'task' in vk and 'subtask' in vk:
            task, subtask, host, function = [vk['task'], vk['subtask'], vk['host'], vk['function']]
            if subtask not in combination[task]:
                combination[task][subtask] = {host: [function]}
            else:
                if host not in combination[task][subtask]:
                    combination[task][subtask][host] = [function]
                else:
                    combination[task][subtask][host].append(function)
    return tasks, combination




"""
The cheml.initialization module includes (please click on links adjacent to function names for more information):
    - XYZreader: :func:`~cheml.initialization.XYZreader`
    - ConvertFile: :func:`~cheml.initialization.ConvertFile`
    - Split: :func:`~cheml.initialization.Split`
    - SaveFile: :func:`~cheml.initialization.SaveFile`

"""

from .initialization import Split

from .initialization import XYZreader
from .initialization import SaveFile
# from .initialization import StoreFile

from .initialization import ConvertFile
# from .data import Trimmer
# from .data import Uniformer

__all__ = [
    'XYZreader',
    'ConvertFile',
    'Split',
    'SaveFile',
]



from .containers import Input, Output, Parameter, req, regression_types, cv_classes

class Binarizer(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'Binarizer'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Binarizer.html#sklearn.preprocessing.Binarizer"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's Binarizer class", ("<class 'sklearn.preprocessing.data.Binarizer'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's Binarizer class", ("<class 'sklearn.preprocessing.data.Binarizer'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        threshold = Parameter('threshold', 0.0)
        copy = Parameter('copy', True)


class PolynomialFeatures(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'PolynomialFeatures'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PolynomialFeatures.html#sklearn.preprocessing.PolynomialFeatures"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's PolynomialFeatures class", ("<class 'sklearn.preprocessing.data.PolynomialFeatures'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's PolynomialFeatures class", ("<class 'sklearn.preprocessing.data.PolynomialFeatures'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        include_bias = Parameter('include_bias', True)
        interaction_only = Parameter('interaction_only', False)
        degree = Parameter('degree', 2)
class OneHotEncoder(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'OneHotEncoder'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html#sklearn.preprocessing.OneHotEncoder"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's OneHotEncoder class", ("<class 'sklearn.preprocessing.data.OneHotEncoder'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's OneHotEncoder class", ("<class 'sklearn.preprocessing.data.OneHotEncoder'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        dtype = Parameter('dtype', "<type'numpy.float64'>")
        n_values = Parameter('n_values', 'auto')
        sparse = Parameter('sparse', True)
        categorical_features = Parameter('categorical_features', 'all')
        handle_unknown = Parameter('handle_unknown', 'error')
class Imputer(object):
    task = 'Prepare'
    subtask = 'data cleaning'
    host = 'sklearn'
    function = 'Imputer'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Imputer.html#sklearn.preprocessing.Imputer"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's Imputer class", ("<class 'sklearn.preprocessing.imputation.Imputer'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's Imputer class", ("<class 'sklearn.preprocessing.imputation.Imputer'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        missing_values = Parameter('missing_values', "NaN")
        copy = Parameter('copy', True)
        verbose = Parameter('verbose', 0)
        strategy = Parameter('strategy', 'mean')
        axis = Parameter('axis', 0)

class KernelPCA(object):
    task = 'Prepare'
    subtask = 'feature transformation'
    host = 'sklearn'
    function = 'KernelPCA'
    modules = ('sklearn','decomposition')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.KernelPCA.html#sklearn.decomposition.KernelPCA"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's KernelPCA class", ("<class 'sklearn.decomposition.kernel_pca.KernelPCA'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's KernelPCA class", ("<class 'sklearn.decomposition.kernel_pca.KernelPCA'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', False,'Boolean',
                        description = "Always False, the header of input dataframe is not equivalent with the transformed dataframe",
                        options = (False))
    class FParameters:
        fit_inverse_transform = Parameter('fit_inverse_transform', False)
        kernel = Parameter('kernel', 'linear')
        n_jobs = Parameter('n_jobs', 1)
        remove_zero_eig = Parameter('remove_zero_eig', False)
        degree = Parameter('degree', 3)
        max_iter = Parameter('max_iter', None)
        kernel_params = Parameter('kernel_params', None)
        random_state = Parameter('random_state', None)
        n_components = Parameter('n_components', None)
        eigen_solver = Parameter('eigen_solver', 'auto')
        tol = Parameter('tol', 0)
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
        coef0 = Parameter('coef0', 1)
        gamma = Parameter('gamma', None)
class PCA(object):
    task = 'Prepare'
    subtask = 'feature transformation'
    host = 'sklearn'
    function = 'PCA'
    modules = ('sklearn','decomposition')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html#sklearn.decomposition.PCA"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's PCA class", ("<class 'sklearn.decomposition.pca.PCA'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's PCA class", ("<class 'sklearn.decomposition.pca.PCA'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', False,'Boolean',
                        description = "Always False, the header of input dataframe is not equivalent with the transformed dataframe",
                        options = (False))
    class FParameters:
        iterated_power = Parameter('iterated_power', 'auto')
        random_state = Parameter('random_state', None)
        whiten = Parameter('whiten', False)
        n_components = Parameter('n_components', None)
        tol = Parameter('tol', 0.0)
        copy = Parameter('copy', True)
        svd_solver = Parameter('svd_solver', 'auto')

class Normalizer(object):
    task = 'Prepare'
    subtask = 'scaling'
    host = 'sklearn'
    function = 'Normalizer'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Normalizer.html#sklearn.preprocessing.Normalizer"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's Normalizer class", ("<class 'sklearn.preprocessing.data.Normalizer'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's Normalizer class", ("<class 'sklearn.preprocessing.data.Normalizer'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api None: only make a new api ",
                        options = ('fit_transform', 'transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        copy = Parameter('copy', True)
        norm = Parameter('norm', 'l2')
class StandardScaler(object):
    task = 'Prepare'
    subtask = 'scaling'
    host = 'sklearn'
    function = 'StandardScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html#sklearn.preprocessing.StandardScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's StandardScaler class", ("<class 'sklearn.preprocessing.data.StandardScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's StandardScaler class", ("<class 'sklearn.preprocessing.data.StandardScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        copy = Parameter('copy', True)
        with_mean = Parameter('with_mean', True)
        with_std = Parameter('with_std', True)
class MinMaxScaler(object):
    task = 'Prepare'
    subtask = 'scaling'
    host = 'sklearn'
    function = 'MinMaxScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html#sklearn.preprocessing.MinMaxScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's MinMaxScaler class", ("<class 'sklearn.preprocessing.data.MinMaxScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's MinMaxScaler class", ("<class 'sklearn.preprocessing.data.MinMaxScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        copy = Parameter('copy', True)
        feature_range = Parameter('feature_range', (0, 1))
class MaxAbsScaler(object):
    task = 'Prepare'
    subtask = 'scaling'
    host = 'sklearn'
    function = 'MaxAbsScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MaxAbsScaler.html#sklearn.preprocessing.MaxAbsScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's MaxAbsScaler class", ("<class 'sklearn.preprocessing.data.MaxAbsScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's MaxAbsScaler class", ("<class 'sklearn.preprocessing.data.MaxAbsScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        copy = Parameter('copy', True)
class RobustScaler(object):
    task = 'Prepare'
    subtask = 'scaling'
    host = 'sklearn'
    function = 'RobustScaler'
    modules = ('sklearn','preprocessing')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html#sklearn.preprocessing.RobustScaler"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Input("api","instance of scikit-learn's RobustScaler class", ("<class 'sklearn.preprocessing.data.RobustScaler'>",))
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's RobustScaler class", ("<class 'sklearn.preprocessing.data.RobustScaler'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                        options = ('fit_transform', 'transform', 'inverse_transform', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        with_centering = Parameter('with_centering', True)
        copy = Parameter('copy', True)
        with_scaling = Parameter('with_scaling', True)
        quantile_range = Parameter('quantile_range', (25.0, 75.0))

class ShuffleSplit(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'ShuffleSplit'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.ShuffleSplit.html#sklearn.model_selection.ShuffleSplit"

    class Inputs:
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
        api = Output("api","instance of scikit-learn's ShuffleSplit class", ("<class 'sklearn.model_selection._split.ShuffleSplit'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        n_splits = Parameter('n_splits', 10)
        test_size = Parameter('test_size', 'default')
        train_size = Parameter('train_size', None)
        random_state = Parameter('random_state', None)

class StratifiedShuffleSplit(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'StratifiedShuffleSplit'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.StratifiedShuffleSplit.html#sklearn.model_selection.StratifiedShuffleSplit"

    class Inputs:
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
        api = Output("api","instance of scikit-learn's StratifiedShuffleSplit class", ("<class 'sklearn.model_selection._split.StratifiedShuffleSplit'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        n_splits = Parameter('n_splits', 10)
        test_size = Parameter('test_size', 'default')
        train_size = Parameter('train_size', None)
        random_state = Parameter('random_state', None)

class train_test_split(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'train_test_split'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        dfx_test = Output("dfx_test","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfy_train = Output("dfy_train","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfy_test = Output("dfy_test","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx_train = Output("dfx_train","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        test_size = Parameter('test_size', 0.25)
        train_size = Parameter('train_size', None)
        random_state = Parameter('random_state', None)
        shuffle = Parameter('shuffle', True)
        stratify = Parameter('stratify', None)

class KFold(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'KFold'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html"

    class Inputs:
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's KFold class", ("<class 'sklearn.model_selection._split.KFold'>",))
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        n_splits = Parameter('n_splits', 3)
        random_state = Parameter('random_state', None)
        shuffle = Parameter('shuffle', False)

class LeaveOneOut(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'LeaveOneOut'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.LeaveOneOut.html"

    class Inputs:
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LeaveOneOut class", ("<class 'sklearn.model_selection._split.LeaveOneOut'>",))
        fold_gen = Output("fold_gen","Generator of indices to split data into training and test set", ("<type 'generator'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('split', None))
    class FParameters:
        pass

class ARDRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'ARDRegression'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ARDRegression.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's ARDRegression class", ("<class 'sklearn.linear_model.bayes.ARDRegression'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's ARDRegression class", ("<class 'sklearn.linear_model.bayes.ARDRegression'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        n_iter = Parameter('n_iter', 300)
        verbose = Parameter('verbose', False)
        lambda_1 = Parameter('lambda_1', 1e-06)
        lambda_2 = Parameter('lambda_2', 1e-06)
        fit_intercept = Parameter('fit_intercept', True)
        threshold_lambda = Parameter('threshold_lambda', 10000.0)
        alpha_2 = Parameter('alpha_2', 1e-06)
        tol = Parameter('tol', 0.001)
        alpha_1 = Parameter('alpha_1', 1e-06)
        copy_X = Parameter('copy_X', True)
        compute_score = Parameter('compute_score', False)
class BayesianRidge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'BayesianRidge'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's BayesianRidge class", ("<class 'sklearn.linear_model.bayes.BayesianRidge'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's BayesianRidge class", ("<class 'sklearn.linear_model.bayes.BayesianRidge'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        n_iter = Parameter('n_iter', 300)
        verbose = Parameter('verbose', False)
        lambda_1 = Parameter('lambda_1', 1e-06)
        lambda_2 = Parameter('lambda_2', 1e-06)
        fit_intercept = Parameter('fit_intercept', True)
        alpha_2 = Parameter('alpha_2', 1e-06)
        tol = Parameter('tol', 0.001)
        alpha_1 = Parameter('alpha_1', 1e-06)
        copy_X = Parameter('copy_X', True)
        compute_score = Parameter('compute_score', False)
class ElasticNet(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'ElasticNet'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's ElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's ElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        warm_start = Parameter('warm_start', False)
        selection = Parameter('selection', 'cyclic')
        fit_intercept = Parameter('fit_intercept', True)
        l1_ratio = Parameter('l1_ratio', 0.5)
        max_iter = Parameter('max_iter', 1000)
        precompute = Parameter('precompute', False)
        random_state = Parameter('random_state', None)
        tol = Parameter('tol', 0.0001)
        positive = Parameter('positive', False)
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
class KernelRidge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'KernelRidge'
    modules = ('sklearn','kernel_ridge')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.kernel_ridge.KernelRidge.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's KernelRidge class", ("<class 'sklearn.kernel_ridge.KernelRidge'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's KernelRidge class", ("<class 'sklearn.kernel_ridge.KernelRidge'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        kernel = Parameter('kernel', 'linear')
        degree = Parameter('degree', 3)
        kernel_params = Parameter('kernel_params', None)
        alpha = Parameter('alpha', 1)
        coef0 = Parameter('coef0', 1)
        gamma = Parameter('gamma', None)
class Lars(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Lars'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lars.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's Lars class", ("<class 'sklearn.linear_model.least_angle.Lars'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's Lars class", ("<class 'sklearn.linear_model.least_angle.Lars'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        n_nonzero_coefs = Parameter('n_nonzero_coefs', 500)
        normalize = Parameter('normalize', True)
        fit_path = Parameter('fit_path', True)
        fit_intercept = Parameter('fit_intercept', True)
        positive = Parameter('positive', False)
        eps = Parameter('eps', 2.22044604925e-16)
        precompute = Parameter('precompute', 'auto')
        copy_X = Parameter('copy_X', True)
        verbose = Parameter('verbose', False)
class Lasso(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Lasso'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's Lasso class", ("<class 'sklearn.linear_model.coordinate_descent.Lasso'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's Lasso class", ("<class 'sklearn.linear_model.coordinate_descent.Lasso'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        warm_start = Parameter('warm_start', False)
        selection = Parameter('selection', 'cyclic')
        fit_intercept = Parameter('fit_intercept', True)
        positive = Parameter('positive', False)
        max_iter = Parameter('max_iter', 1000)
        precompute = Parameter('precompute', False)
        random_state = Parameter('random_state', None)
        tol = Parameter('tol', 0.0001)
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
class LassoLars(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LassoLars'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LassoLars class", ("<class 'sklearn.linear_model.least_angle.LassoLars'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LassoLars class", ("<class 'sklearn.linear_model.least_angle.LassoLars'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', True)
        fit_path = Parameter('fit_path', True)
        fit_intercept = Parameter('fit_intercept', True)
        positive = Parameter('positive', False)
        max_iter = Parameter('max_iter', 500)
        eps = Parameter('eps', 2.22044604925e-16)
        precompute = Parameter('precompute', 'auto')
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
        verbose = Parameter('verbose', False)
class LinearRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LinearRegression'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LinearRegression class", ("<class 'sklearn.linear_model.base.LinearRegression'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LinearRegression class", ("<class 'sklearn.linear_model.base.LinearRegression'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        copy_X = Parameter('copy_X', True)
        normalize = Parameter('normalize', False)
        n_jobs = Parameter('n_jobs', 1)
        fit_intercept = Parameter('fit_intercept', True)
class LinearSVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LinearSVR'
    modules = ('sklearn','svm')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVR.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LinearSVR class", ("<class 'sklearn.svm.classes.LinearSVR'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LinearSVR class", ("<class 'sklearn.svm.classes.LinearSVR'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        loss = Parameter('loss', 'epsilon_insensitive')
        C = Parameter('C', 1.0)
        intercept_scaling = Parameter('intercept_scaling', 1.0)
        fit_intercept = Parameter('fit_intercept', True)
        epsilon = Parameter('epsilon', 0.0)
        max_iter = Parameter('max_iter', 1000)
        random_state = Parameter('random_state', None)
        dual = Parameter('dual', True)
        tol = Parameter('tol', 0.0001)
        verbose = Parameter('verbose', 0)
class LogisticRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LogisticRegression'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's LogisticRegression class", ("<class 'sklearn.linear_model.logistic.LogisticRegression'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's LogisticRegression class", ("<class 'sklearn.linear_model.logistic.LogisticRegression'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        warm_start = Parameter('warm_start', False)
        C = Parameter('C', 1.0)
        n_jobs = Parameter('n_jobs', 1)
        verbose = Parameter('verbose', 0)
        intercept_scaling = Parameter('intercept_scaling', 1)
        fit_intercept = Parameter('fit_intercept', True)
        max_iter = Parameter('max_iter', 100)
        penalty = Parameter('penalty', 'l2')
        multi_class = Parameter('multi_class', 'ovr')
        random_state = Parameter('random_state', None)
        dual = Parameter('dual', False)
        tol = Parameter('tol', 0.0001)
        solver = Parameter('solver', 'liblinear')
        class_weight = Parameter('class_weight', None)
class MLPRegressor(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MLPRegressor'
    modules = ('sklearn','neural_network')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html#sklearn.neural_network.MLPRegressor"

    class Inputs:
        api = Input("api","instance of scikit-learn's MLPRegressor class", ("<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's MLPRegressor class", ("<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        beta_1 = Parameter('beta_1', 0.9)
        warm_start = Parameter('warm_start', False)
        beta_2 = Parameter('beta_2', 0.999)
        shuffle = Parameter('shuffle', True)
        verbose = Parameter('verbose', False)
        nesterovs_momentum = Parameter('nesterovs_momentum', True)
        hidden_layer_sizes = Parameter('hidden_layer_sizes', (100,))
        epsilon = Parameter('epsilon', 1e-08)
        activation = Parameter('activation', 'relu')
        max_iter = Parameter('max_iter', 200)
        batch_size = Parameter('batch_size', 'auto')
        power_t = Parameter('power_t', 0.5)
        random_state = Parameter('random_state', None)
        learning_rate_init = Parameter('learning_rate_init', 0.001)
        tol = Parameter('tol', 0.0001)
        validation_fraction = Parameter('validation_fraction', 0.1)
        alpha = Parameter('alpha', 0.0001)
        solver = Parameter('solver', 'adam')
        momentum = Parameter('momentum', 0.9)
        learning_rate = Parameter('learning_rate', 'constant')
        early_stopping = Parameter('early_stopping', False)
class MultiTaskElasticNet(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MultiTaskElasticNet'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskElasticNet.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's MultiTaskElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's MultiTaskElasticNet class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        warm_start = Parameter('warm_start', False)
        selection = Parameter('selection', 'cyclic')
        fit_intercept = Parameter('fit_intercept', True)
        l1_ratio = Parameter('l1_ratio', 0.5)
        max_iter = Parameter('max_iter', 1000)
        random_state = Parameter('random_state', None)
        tol = Parameter('tol', 0.0001)
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
class MultiTaskLasso(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MultiTaskLasso'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskLasso.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's MultiTaskLasso class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's MultiTaskLasso class", ("<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        warm_start = Parameter('warm_start', False)
        selection = Parameter('selection', 'cyclic')
        fit_intercept = Parameter('fit_intercept', True)
        max_iter = Parameter('max_iter', 1000)
        random_state = Parameter('random_state', None)
        tol = Parameter('tol', 0.0001)
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
class NuSVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'NuSVR'
    modules = ('sklearn','svm')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVR.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's NuSVR class", ("<class 'sklearn.svm.classes.NuSVR'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's NuSVR class", ("<class 'sklearn.svm.classes.NuSVR'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        kernel = Parameter('kernel', 'rbf')
        C = Parameter('C', 1.0)
        verbose = Parameter('verbose', False)
        degree = Parameter('degree', 3)
        shrinking = Parameter('shrinking', True)
        max_iter = Parameter('max_iter', -1)
        tol = Parameter('tol', 0.001)
        cache_size = Parameter('cache_size', 200)
        coef0 = Parameter('coef0', 0.0)
        nu = Parameter('nu', 0.5)
        gamma = Parameter('gamma', 'auto')
class Ridge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Ridge'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's Ridge class", ("<class 'sklearn.linear_model.ridge.Ridge'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's Ridge class", ("<class 'sklearn.linear_model.ridge.Ridge'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        normalize = Parameter('normalize', False)
        fit_intercept = Parameter('fit_intercept', True)
        max_iter = Parameter('max_iter', None)
        random_state = Parameter('random_state', None)
        tol = Parameter('tol', 0.001)
        copy_X = Parameter('copy_X', True)
        alpha = Parameter('alpha', 1.0)
        solver = Parameter('solver', 'auto')
class SGDRegressor(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'SGDRegressor'
    modules = ('sklearn','linear_model')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDRegressor.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's SGDRegressor class", ("<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's SGDRegressor class", ("<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        warm_start = Parameter('warm_start', False)
        loss = Parameter('loss', 'squared_loss')
        eta0 = Parameter('eta0', 0.01)
        verbose = Parameter('verbose', 0)
        shuffle = Parameter('shuffle', True)
        fit_intercept = Parameter('fit_intercept', True)
        l1_ratio = Parameter('l1_ratio', 0.15)
        average = Parameter('average', False)
        n_iter = Parameter('n_iter', 5)
        penalty = Parameter('penalty', 'l2')
        power_t = Parameter('power_t', 0.25)
        random_state = Parameter('random_state', None)
        epsilon = Parameter('epsilon', 0.1)
        alpha = Parameter('alpha', 0.0001)
        learning_rate = Parameter('learning_rate', 'invscaling')
class SVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'SVR'
    modules = ('sklearn','svm')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html"

    class Inputs:
        api = Input("api","instance of scikit-learn's SVR class", ("<class 'sklearn.svm.classes.SVR'>",))
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        api = Output("api","instance of scikit-learn's SVR class", ("<class 'sklearn.svm.classes.SVR'>",))
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        func_method = Parameter('func_method','None','string',
                        description = "",
                        options = ('fit', 'predict', None))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        kernel = Parameter('kernel', 'rbf')
        C = Parameter('C', 1.0)
        verbose = Parameter('verbose', False)
        degree = Parameter('degree', 3)
        epsilon = Parameter('epsilon', 0.1)
        shrinking = Parameter('shrinking', True)
        max_iter = Parameter('max_iter', -1)
        tol = Parameter('tol', 0.001)
        cache_size = Parameter('cache_size', 200)
        coef0 = Parameter('coef0', 0.0)
        gamma = Parameter('gamma', 'auto')

class GridSearchCV(object):
    task = 'Search'
    subtask = 'grid'
    host = 'sklearn'
    function = 'GridSearchCV'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", tuple(["<type 'str'>"]+list(regression_types())))
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
        cv = Input("cv", "instance of scikit-learn's cross validation generator or instance object", tuple(["<type 'generator'>"]+list(cv_classes())))

    class Outputs:
        cv_results_ = Output("cv_results_","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        api = Output("api","instance of scikit-learn's GridSearchCV class", ("<class 'sklearn.grid_search.GridSearchCV'>",))
        best_estimator_ = Output("best_estimator_","instance of a machine learning class", regression_types())
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        scoring = Parameter('scoring', None)
        n_jobs = Parameter('n_jobs', 1)
        verbose = Parameter('verbose', 0)
        fit_params = Parameter('fit_params', None)
        refit = Parameter('refit', True)
        return_train_score = Parameter('return_train_score', True)
        iid = Parameter('iid', True)
        estimator = Parameter('estimator', '@estimator', required=True)
        error_score = Parameter('error_score','raise')
        pre_dispatch = Parameter('pre_dispatch', '2 * n_jobs')
        param_grid = Parameter('param_grid', {}, required=True)
        cv = Parameter('cv', None)

class cross_val_predict(object):
    task = 'Search'
    subtask = 'validate'
    host = 'sklearn'
    function = 'cross_val_predict'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_val_predict.html#sklearn.model_selection.cross_val_predict"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", regression_types())
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
        cv = Input("cv","cross-validation generator or instance object", tuple(["<type 'generator'>"]+list(cv_classes())))
    class Outputs:
        dfy_predict = Output("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        estimator = Parameter('estimator', '@estimator', required=True)
        X = Parameter('X', '@dfx', required=True)
        y = Parameter('y', None)
        groups = Parameter('groups', None)
        cv = Parameter('cv', None)
        n_jobs = Parameter('n_jobs', 1)
        verbose = Parameter('verbose', 0)
        fit_params = Parameter('fit_params', None)
        pre_dispatch = Parameter('pre_dispatch', '2 * n_jobs')
        method = Parameter('method', 'predict')
class cross_val_score(object):
    task = 'Search'
    subtask = 'validate'
    host = 'sklearn'
    function = 'cross_val_score'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_val_score.html#sklearn.model_selection.cross_val_score"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", regression_types())
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
        cv = Input("cv", "cross-validation generator or instance object", tuple(["<type 'generator'>"]+list(cv_classes())))

    class Outputs:
        scores = Output("scores","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        estimator = Parameter('estimator', '@estimator', required=True)
        X = Parameter('X', '@dfx', required=True)
        y = Parameter('y', None)
        groups = Parameter('groups', None)
        scoring = Parameter('scoring', None)
        cv = Parameter('cv', None)
        n_jobs = Parameter('n_jobs', 1)
        verbose = Parameter('verbose', 0)
        fit_params = Parameter('fit_params', None)
        pre_dispatch = Parameter('pre_dispatch', '2 * n_jobs')
class learning_curve(object):
    task = 'Search'
    subtask = 'grid'
    host = 'sklearn'
    function = 'learning_curve'
    modules = ('sklearn','model_selection')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.learning_curve.html#sklearn.model_selection.learning_curve"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfx = Input("dfx","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        estimator = Input("estimator","instance of a machine learning class", regression_types())
        scorer = Input("scorer","instance of scikit-learn's make_scorer class", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))
        cv = Input("cv","instance of scikit-learn's cross validation generator or instance object",
                   tuple(["<type 'generator'>"] + list(cv_classes())))
    class Outputs:
        train_sizes_abs = Output("train_sizes_abs","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        extended_result_ = Output("extended_result_","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        test_scores = Output("test_scores","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        train_scores = Output("train_scores","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class WParameters:
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        estimator = Parameter('estimator', '@estimator', required=True)
        X = Parameter('X', '@dfx', required=True)
        y = Parameter('y', None)
        groups = Parameter('groups', None)
        train_sizes = Parameter('train_sizes', [ 0.1, 0.33, 0.55, 0.78,1.])
        scoring = Parameter('scoring', None)
        cv = Parameter('cv', None)
        exploit_incremental_learning = Parameter('exploit_incremental_learning', False)
        n_jobs = Parameter('n_jobs', 1)
        pre_dispatch = Parameter('pre_dispatch', 'all')
        verbose = Parameter('verbose', 0)
        shuffle = Parameter('shuffle', False)
        random_state = Parameter('random_state', None)
class evaluate_regression(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'sklearn'
    function = 'evaluate_regression'
    modules = ('sklearn','metrics')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/dev/modules/model_evaluation.html#regression-metrics"

    class Inputs:
        dfy = Input("dfy","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        dfy_predict = Input("dfy_predict","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
    class Outputs:
        evaluation_results_ = Output("evaluation_results_","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        evaluator = Output("evaluator","dictionary of metrics and their score function", ("<type 'dict'>",))
    class WParameters:
        r2_score = Parameter('r2_score','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score',
                        description = "",
                        options = (True, False))
        r2_sample_weight = Parameter('r2_sample_weight', 'None',
                                     'http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score',
                                     description="",
                                     options=[])
        multioutput = Parameter('multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score',
                        description = "",
                        options = ('raw_values', 'uniform_average', 'variance_weighted'))

        mean_absolute_error = Parameter('mean_absolute_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error',
                        description = "",
                        options = (True, False))
        mae_multioutput = Parameter('mae_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error',
                        description = "",
                        options = ('raw_values', 'uniform_average'))
        mae_sample_weight = Parameter('mae_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error',
                        description = "",
                        options = [])

        mse_sample_weight = Parameter('mse_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = [])
        rmse_multioutput = Parameter('rmse_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = ('raw_values', 'uniform_average'))
        median_absolute_error = Parameter('median_absolute_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.median_absolute_error.html#sklearn.metrics.median_absolute_error',
                        description = "",
                        options = (True, False))
        rmse_sample_weight = Parameter('rmse_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = [])
        mean_squared_error = Parameter('mean_squared_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = (True, False))
        root_mean_squared_error = Parameter('root_mean_squared_error','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = (True, False))
        explained_variance_score = Parameter('explained_variance_score','False','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score',
                        description = "",
                        options = (True, False))

        ev_sample_weight = Parameter('ev_sample_weight','None','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score',
                        description = "",
                        options = [])
        ev_multioutput = Parameter('ev_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score',
                        description = "",
                        options = ('raw_values', 'uniform_average', 'variance_weighted'))
        mse_multioutput = Parameter('mse_multioutput','uniform_average','http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error',
                        description = "",
                        options = ('raw_values', 'uniform_average'))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        pass
class scorer_regression(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'sklearn'
    function = 'scorer_regression'
    modules = ('sklearn','metrics')
    requirements = (req(1), req(2))
    documentation = "http://scikit-learn.org/0.15/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer"

    class Inputs:
        pass

    class Outputs:
        scorer = Output("scorer","Callable object that returns a scalar score", ("<class 'sklearn.metrics.scorer._PredictScorer'>",))

    class WParameters:
        metric = Parameter('metric','mae',
                        description = "http://scikit-learn.org/dev/modules/model_evaluation.html#regression-metrics",
                        format = "string: 'mae', 'mse', 'r2'",
                        options = ('mae', 'mse', 'r2'))
        track_header = Parameter('track_header', True,'Boolean',
                        description = "if True, the input dataframe's header will be transformed to the output dataframe",
                        options = (True, False))
    class FParameters:
        greater_is_better = Parameter('greater_is_better', True)
        needs_proba = Parameter('needs_proba', False)
        needs_threshold = Parameter('needs_threshold', False)
        kwargs = Parameter('kwargs', {}, format = 'dictionary')






import warnings
import pandas as pd
import numpy as np
import copy


from .syntax import Preprocessor, Regressor, Evaluator
from ..base import BASE
from ...utils import value


##################################################################### 2 Prepare Data

class automatic_run(BASE):
    def fit(self):
        self.Fit_sklearn()
        # tokens = ['df','dfy_predict']
        # for token in tokens:
        #     if token in self.outputs:
        #         if self.outputs[token].value is not None:
        #             if self.header:
        #                 df = self.outputs[token].value
        #                 df = pd.DataFrame(df, columns=self.inputs[token[0:3]].value.columns)
        #                 self.outputs[token].value = df
        #             else:
        #                 self.outputs[token].value = pd.DataFrame(self.outputs[token].value)
        self.Send()
        # delete all inputs
        del self.inputs


# feature representation
class PolynomialFeatures(BASE):
    def fit(self):
        self.Fit_sklearn()
        # headers from class method: get_feature_names
        if self.method in ['fit_transform','transform']:
            if self.header:
                df = self.outputs['df'].value
                df = pd.DataFrame(df, columns = self.outputs['api'].value.get_feature_names())
                self.outputs['df'].value = df
            else:
                df = self.outputs['df'].value
                df = pd.DataFrame(df)
                self.outputs['df'].value = df

        self.Send()
        # delete all inputs
        del self.inputs


class OneHotEncoder(BASE):
    def fit(self):
        self.Fit_sklearn()
        # .toarray() is requied
        if self.method in ['fit_transform','transform']:
            df = self.outputs['df'].value.toarray()
            df = pd.DataFrame(df, columns = self.inputs['df'].value.columns)
            self.outputs['df'].value = df
        self.Send()
        # delete all inputs
        del self.inputs


# basic oprator

class Imputer(BASE):
    def fit(self):
        self.Fit_sklearn()
        # headers from input df and based on the class attribute: statistics_
        if self.method in ['fit_transform', 'transform']:
            if self.header:
                df = self.outputs['df'].value
                api = self.outputs['api'].value
                stats = api.statistics_
                NOTnan_ind = [i for i, val in enumerate(stats) if not np.isnan(val)]
                df = pd.DataFrame(df, columns=self.inputs['df'].value.columns[NOTnan_ind])
                self.outputs['df'].value = df
            else:
                df = self.outputs['df'].value
                df = pd.DataFrame(df)
                self.outputs['df'].value = df
        self.Send()
        # delete all inputs
        del self.inputs

# scaler

# feature selector

# feature transformer


# splitter

class train_test_split(BASE):
    def fit(self):
        # step1: check inputs
        self.required('dfx', req=True)
        dfx = self.inputs['dfx'].value
        dfy = self.inputs['dfy'].value

        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()

        # step3: check the dimension of input data frame
        dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='df')
        # step4: import module and make APIs
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, self.Function)
            if dfy is None:
                tts_out = F(dfx, **self.parameters)
                dfx_train, dfx_test = tts_out
                self.set_value('dfx_train', pd.DataFrame(dfx_train,columns=dfx.columns))
                self.set_value('dfx_test',pd.DataFrame(dfx_test,columns=dfx.columns))
            else:
                dfy, _ = self.data_check('dfy', dfy, ndim=1, n0=dfx.shape[0], n1=None, format_out='df')
                tts_out = F(dfx, dfy, **self.parameters)
                dfx_train, dfx_test, dfy_train, dfy_test = tts_out
                self.set_value('dfx_train', pd.DataFrame(dfx_train,columns=dfx.columns))
                self.set_value('dfx_test', pd.DataFrame(dfx_test,columns=dfx.columns))
                self.set_value('dfy_train', pd.DataFrame(dfy_train,columns=dfy.columns))
                self.set_value('dfy_test', pd.DataFrame(dfy_test,columns=dfy.columns))
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + err.message
            raise TypeError(msg)


        # send and delete inputs
        self.Send()
        del self.inputs
        del dfx
        del dfy

class KFold(BASE):
    def fit(self):
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None
        if method not in self.metadata.WParameters.func_method.options:
            msg = "@Task #%i(%s): The method '%s' is not available for the function '%s'." % (
                self.iblock, self.Task,method,self.Function)
            raise NameError(msg)
        else:
            if method == None:
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.set_value('api', api)
            elif method == 'split':
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.required('dfx', req=True)
                dfx = self.inputs['dfx'].value
                fold_gen = api.split(dfx)
                self.set_value('api', api)
                self.set_value('fold_gen', fold_gen)

        self.Send()
        del self.inputs

class LeaveOneOut(BASE):
    def fit(self):
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None
        if method not in self.metadata.WParameters.func_method.options:
            msg = "@Task #%i(%s): The method '%s' is not available for the function '%s'." % (
                self.iblock, self.Task,method,self.Function)
            raise NameError(msg)
        else:
            if method == None:
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.set_value('api', api)
            elif method == 'split':
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.required('dfx', req=True)
                dfx = self.inputs['dfx'].value
                fold_gen = api.split(dfx)
                self.set_value('api', api)
                self.set_value('fold_gen', fold_gen)

        self.Send()
        del self.inputs

class ShuffleSplit(BASE):
    def fit(self):
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None
        if method not in self.metadata.WParameters.func_method.options:
            msg = "@Task #%i(%s): The method '%s' is not available for the function '%s'." % (
                self.iblock, self.Task,method,self.Function)
            raise NameError(msg)
        else:
            if method is None:
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.set_value('api', api)
            elif method is 'split':
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.required('dfx', req=True)
                dfx = self.inputs['dfx'].value
                fold_gen = api.split(dfx)
                self.set_value('api', api)
                self.set_value('fold_gen', fold_gen)

        self.Send()
        del self.inputs

class StratifiedShuffleSplit(BASE):
    def fit(self):
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None
        if method not in self.metadata.WParameters.func_method.options:
            msg = "@Task #%i(%s): The method '%s' is not available for the function '%s'." % (
                self.iblock, self.Task,method,self.Function)
            raise NameError(msg)
        else:
            if method is None:
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.set_value('api', api)
            elif method is 'split':
                try:
                    exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
                    submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
                    F = getattr(submodule, self.Function)
                    api = F(**self.parameters)
                except Exception as err:
                    msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
                    raise TypeError(msg)
                self.required('dfx', req=True)
                dfx = self.inputs['dfx'].value
                fold_gen = api.split(dfx)
                self.set_value('api', api)
                self.set_value('fold_gen', fold_gen)

        self.Send()
        del self.inputs


##################################################################### 3 Define Model

# Regression


##################################################################### 4 Search

class GridSearchCV(BASE):
    def fit(self):
        # step1: check inputs
        self.required('dfx', req=True)
        self.required('estimator', req=True)
        dfx = self.inputs['dfx'].value
        dfy = self.inputs['dfy'].value
        estimator = self.inputs['estimator'].value
        self.parameters['estimator'] = estimator

        # step2: check parameters
        # Note: estimator is a required parameter and can be received from the input stream
        # Note: scorer is not a required parameter but can be received from the input stream
        self.paramFROMinput()

        # step3: check the dimension of input data frame
        dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
        if dfy is not None:
            dfy, _ = self.data_check('dfy', dfy, ndim=1, n0=dfx.shape[0], n1=None, format_out='ar')

        print( dfx.shape)
        print( dfy.shape)

        # step4: import module and make API
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, self.Function)
            api = F(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + err.message
            raise TypeError(msg)

        # step5: process
        api.fit(dfx, dfy)

        # step6: send
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'best_estimator_':
                if api.get_params()['refit']:
                    best_estimator_ = copy.deepcopy(api.best_estimator_)
                else:
                    best_estimator_ = copy.deepcopy(self.parameters['estimator'])
                    best_estimator_.set_params(**api.best_params_)
                    # best_estimator_.fit(dfx,dfy)
                self.set_value('best_estimator_', best_estimator_)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'cv_results_':
                self.set_value('cv_results_', pd.DataFrame(api.cv_results_))
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'api':
                self.set_value('api',api)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete inputs
        del self.inputs
        del dfx
        del dfy

class cross_val_score(BASE):
    def fit(self):
        # step1: check inputs
        self.required('dfx', req=True)
        self.required('estimator', req=True)
        dfx = self.inputs['dfx'].value
        dfy = self.inputs['dfy'].value
        estimator = self.inputs['estimator'].value

        # step2: check parameters
        # Note: estimator is a required parameter and can be received from the input stream
        # Note: scorer is not a required parameter but can be received from the input stream
        self.paramFROMinput()
        if 'estimator' not in self.parameters:
            self.parameters['estimator'] = estimator
        if 'X' not in self.parameters:
            self.parameters['X'] = dfx


        # step3: check the dimension of input data frame
        dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
        if dfy is not None:
            dfy, _ = self.data_check('dfy', dfy, ndim=1, n0=dfx.shape[0], n1=None, format_out='ar')

        # step4: import module and make APIs
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, self.Function)
            scores = F(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + err.message
            raise TypeError(msg)

        # step6: set outputs
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token == 'scores':
                self.set_value(token, pd.DataFrame(scores))
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            else:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)

        # step7: delete all inputs from memory
        del self.inputs
        del dfx
        del dfy

class cross_val_predict(BASE):
    def fit(self):
        # step1: check inputs
        self.required('dfx', req=True)
        self.required('estimator', req=True)
        dfx = self.inputs['dfx'].value
        dfy = self.inputs['dfy'].value
        estimator = self.inputs['estimator'].value
        self.parameters['estimator'] = estimator
        # step2: check parameters
        # Note: estimator is a required parameter and can be received from the input stream
        # Note: scorer is not a required parameter but can be received from the input stream
        self.paramFROMinput()

        # step3: check the dimension of input data frame
        dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
        if dfy is not None:
            dfy, _ = self.data_check('dfy', dfy, ndim=1, n0=dfx.shape[0], n1=None, format_out='ar')

        # step4: import module and make APIs
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, self.Function)
            dfy_pred = F(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + err.message
            raise TypeError(msg)

        # step6: set outputs
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token == 'dfy_predict':
                self.set_value(token, pd.DataFrame(dfy_pred))
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            else:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)

        # step7: delete all inputs from memory
        del self.inputs
        del dfx
        del dfy

class learning_curve(BASE):
    def fit(self):
        # step1: check inputs
        self.required('dfx', req=True)
        self.required('estimator', req=True)
        dfx = self.inputs['dfx'].value
        dfy = self.inputs['dfy'].value
        estimator = self.inputs['estimator'].value
        self.parameters['estimator'] = estimator

        # step2: check parameters
        # Note: estimator is a required parameter and can be received from the input stream
        # Note: scorer is not a required parameter but can be received from the input stream
        self.paramFROMinput()
        if type(self.parameters['train_sizes']) is str:
            self.parameters['train_sizes'] = eval(self.parameters['train_sizes'])

        # step3: check the dimension of input data frame
        dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
        if dfy is not None:
            dfy, _ = self.data_check('dfy', dfy, ndim=1, n0=dfx.shape[0], n1=None, format_out='ar')

        # step4: import module and make APIs
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, self.Function)
            train_sizes_abs, train_scores, test_scores = F(X=dfx,y=dfy,**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + err.message
            raise TypeError(msg)

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token == 'train_sizes_abs':
                self.set_value(token, train_sizes_abs)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'train_scores':
                self.set_value(token, train_scores)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'test_scores':
                self.set_value(token, test_scores)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'extended_result_':
                train_scores_mean = np.mean(train_scores, axis=1)
                train_scores_std = np.std(train_scores, axis=1)
                test_scores_mean = np.mean(test_scores, axis=1)
                test_scores_std = np.std(test_scores, axis=1)
                data = [train_sizes_abs, train_scores_mean, train_scores_std, test_scores_mean, test_scores_std]
                cols = ['train_sizes', 'train_scores_mean', 'train_scores_std', 'test_scores_mean', 'test_scores_std']
                extended_result_ = pd.DataFrame()
                for i, col in enumerate(cols):
                    extended_result_[col] = data[i]

                self.set_value(token, extended_result_)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            else:
                msg = "@Task #%i(%s): non valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)

        # step7: delete all inputs from memory
        del self.inputs


# Metrics

class evaluate_regression(BASE,Evaluator):
    def fit(self):
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()

        # step4: import module and make APIs
        try:
            self._reg_evaluation_params()
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ err.message
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token == 'evaluator':
                self.set_value(token, self.evaluator)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'evaluation_results_':
                # step1: check inputs
                self.required('dfy', req=True)
                self.required('dfy_predict', req=True)
                dfy = self.inputs['dfy'].value
                dfy_predict = self.inputs['dfy_predict'].value

                # step3: check the dimension of input data frame
                dfy, _ = self.data_check('dfy', dfy, ndim=2, n0=None, n1=None, format_out='df')
                dfy_predict, _ = self.data_check('dfy_predict', dfy_predict, ndim=2, n0=dfy.shape[0], n1=None, format_out='df')

                self._reg_evaluate(dfy, dfy_predict, self.evaluator)
                evaluation_results_ = pd.DataFrame(self.results)
                self.set_value(token, evaluation_results_)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            else:
                msg = "@Task #%i(%s): non valid output token '%s'" % (self.iblock+1, self.Task, token)
                raise NameError(msg)

        #step7: delete all inputs from memory
        del self.inputs

class scorer_regression(BASE):
    def fit(self):
        # step1: check inputs
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        if 'metric' in self.parameters:
            metric = self.parameters.pop('metric')
        else:
            metric = 'mae'
        if 'kwargs' in self.parameters:
            kwargs = self.parameters.pop('kwargs')
            for item in kwargs:
                self.parameters[item] = value(kwargs[item])

        # Todo: add all the metrics for regression

        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], 'make_scorer'))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, 'make_scorer')
            if metric == 'mae':
                from sklearn.metrics import mean_absolute_error
                scorer = F(mean_absolute_error,**self.parameters)
            elif metric =='r2':
                from sklearn.metrics import r2_score
                scorer = F(r2_score,**self.parameters)
            elif metric == 'mse':
                from sklearn.metrics import mean_squared_error
                scorer = F(mean_squared_error, **self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + err.message
            raise TypeError(msg)
        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token == 'scorer':
                self.set_value(token, scorer)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            else:
                msg = "@Task #%i(%s): non valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
        # step7: delete all inputs from memory
        del self.inputs


##################################################################### 5 Mix




from .containers import Input, Output, Parameter, req, regression_types, cv_classes

class read_table(object):
    task = 'Enter'
    subtask = 'table'
    host = 'pandas'
    function = 'read_table'
    modules = ('pandas','')
    requirements = (req(2),)
    documentation = "https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_table.html"

    class Inputs:
        pass
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class WParameters:
        pass
    class FParameters:
        filepath_or_buffer = Parameter('filepath_or_buffer', 'required_required')
        sep = Parameter('sep', "required_required")
        delimiter = Parameter('delimiter', None)
        header = Parameter('header', 'infer')
        names = Parameter('names', None)
        index_col = Parameter('index_col', None)
        usecols = Parameter('usecols', None)
        squeeze = Parameter('squeeze', False)
        prefix = Parameter('prefix', None)
        mangle_dupe_cols = Parameter('mangle_dupe_cols', True)
        dtype = Parameter('dtype', None)
        engine = Parameter('engine', None)
        converters = Parameter('converters', None)
        true_values = Parameter('true_values', None)
        false_values = Parameter('false_values', None)
        skipinitialspace = Parameter('skipinitialspace', False)
        skiprows = Parameter('skiprows', None)
        nrows = Parameter('nrows', None)
        na_values = Parameter('na_values', None)
        keep_default_na = Parameter('keep_default_na', True)
        na_filter = Parameter('na_filter', True)
        verbose = Parameter('verbose', False)
        skip_blank_lines = Parameter('skip_blank_lines', True)
        parse_dates = Parameter('parse_dates', False)
        infer_datetime_format = Parameter('infer_datetime_format', False)
        keep_date_col = Parameter('keep_date_col', False)
        date_parser = Parameter('date_parser', None)
        dayfirst = Parameter('dayfirst', False)
        iterator = Parameter('iterator', False)
        chunksize = Parameter('chunksize', None)
        compression = Parameter('compression', 'infer')
        thousands = Parameter('thousands', None)
        decimal = Parameter('decimal', '.')
        lineterminator = Parameter('lineterminator', None)
        quotechar = Parameter('quotechar', '"')
        quoting = Parameter('quoting', 0)
        escapechar = Parameter('escapechar', None)
        comment = Parameter('comment', None)
        encoding = Parameter('encoding', None)
        dialect = Parameter('dialect', None)
        tupleize_cols = Parameter('tupleize_cols', None)
        error_bad_lines = Parameter('error_bad_lines', True)
        warn_bad_lines = Parameter('warn_bad_lines', True)
        skipfooter = Parameter('skipfooter', 0)
        skip_footer = Parameter('skip_footer', 0)
        doublequote = Parameter('doublequote', True)
        delim_whitespace = Parameter('delim_whitespace', False)
        as_recarray = Parameter('as_recarray', None)
        compact_ints = Parameter('compact_ints', None)
        use_unsigned = Parameter('use_unsigned', None)
        low_memory = Parameter('low_memory', True)
        buffer_lines = Parameter('buffer_lines', None)
        memory_map = Parameter('memory_map', False)
        float_precision = Parameter('float_precision', None)

class read_excel(object):
    task = 'Enter'
    subtask = 'table'
    host = 'pandas'
    function = 'read_excel'
    modules = ('pandas','')
    requirements = (req(2),)
    documentation = "https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_excel.html"

    class Inputs:
        pass
    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class WParameters:
        pass
    class FParameters:
        io = Parameter('io', 'required_required')
        sheet_name = Parameter('sheet_name', 0)
        header = Parameter('header', 0)
        skiprows = Parameter('skiprows', None)
        skip_footer = Parameter('skip_footer', 0)
        index_col = Parameter('index_col', None)
        names = Parameter('names', None)
        usecols = Parameter('usecols', None)
        parse_dates = Parameter('parse_dates', False)
        date_parser = Parameter('date_parser', None)
        na_values = Parameter('na_values', None)
        thousands = Parameter('thousands', None)
        convert_float = Parameter('convert_float', True)
        converters = Parameter('converters', None)
        dtype = Parameter('dtype', None)
        true_values = Parameter('true_values', None)
        false_values = Parameter('false_values', None)
        engine = Parameter('engine', None)
        squeeze = Parameter('squeeze', False)

class corr(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'pandas'
    function = 'corr'
    modules = ('pandas','')
    requirements = (req(2),)
    documentation = "https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.corr.html"

    class Inputs:
        df = Input("df", "pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class WParameters:
        pass
    class FParameters:
        method = Parameter('method', 'pearson')
        min_periods = Parameter('min_periods', 1)

class concat(object):
    task = 'Prepare'
    subtask = 'data manipulation'
    host = 'pandas'
    function = 'concat'
    modules = ('pandas','')
    requirements = (req(2),)
    documentation = "http://pandas.pydata.org/pandas-docs/stable/generated/pandas.concat.html"

    class Inputs:
        df1 = Input("df1","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        df2 = Input("df2","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))
        df3 = Input("df3", "pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class Outputs:
        df = Output("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class WParameters:
        pass
    class FParameters:
        join = Parameter('join', 'outer')
        axis = Parameter('axis', 0)
        join_axes = Parameter('join_axes', None)
        ignore_index = Parameter('ignore_index', False)
        keys = Parameter('keys', None)
        levels = Parameter('levels', None)
        names = Parameter('names', None)
        verify_integrity = Parameter('verify_integrity', False)
        copy = Parameter('copy', True)

class plot(object):
    task = 'Visualize'
    subtask = 'plot'
    host = 'pandas'
    function = 'plot'
    modules = ('pandas','')
    requirements = (req(2),req(7))
    documentation = "https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.plot.html"

    class Inputs:
        df = Input("df","pandas dataframe", ("<class 'pandas.core.frame.DataFrame'>",))

    class Outputs:
        fig = Output("fig","matplotlib figure or axes object",
                      ("<class 'matplotlib.axes._subplots.AxesSubplot'>","<class 'matplotlib.figure.Figure'>"))

    class WParameters:
        pass
    class FParameters:
        x = Parameter('x', None)
        y = Parameter('y', None)
        kind = Parameter('kind', 'line')
        ax = Parameter('ax', None)
        subplots = Parameter('subplots', False)
        sharex = Parameter('sharex', None)
        sharey = Parameter('sharey', False)
        layout = Parameter('layout', None)
        figsize = Parameter('figsize', None)
        use_index = Parameter('use_index', True)
        title = Parameter('title', None)
        grid = Parameter('grid', None)
        legend = Parameter('legend', True)
        style = Parameter('style', None)
        logx = Parameter('logx', False)
        logy = Parameter('logy', False)
        loglog = Parameter('loglog', False)
        xticks = Parameter('xticks', None)
        yticks = Parameter('yticks', None)
        xlim = Parameter('xlim', None)
        ylim = Parameter('ylim', None)
        rot = Parameter('rot', None)
        fontsize = Parameter('fontsize', None)
        colormap = Parameter('colormap', None)
        table = Parameter('table', False)
        yerr = Parameter('yerr', None)
        xerr = Parameter('xerr', None)
        secondary_y = Parameter('secondary_y', False)
        sort_columns = Parameter('sort_columns', False)



from __future__ import print_function

import os
import pandas as pd


from ..base import BASE

##################################################################### 1 Enter Data

# input

class read_table(BASE):
    def fit(self):
        # step1: check inputs
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            df = pd.read_table(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ err.message
            raise TypeError(msg)

        # step5: process
        print( '(rows, columns): ', df.shape)
        # if 'header' in self.parameters and self.parameters['header'] is not None:
        #     print 'headers: ', list(df.columns)

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]


        # step7: delete all inputs from memory
        del self.inputs

class read_excel(BASE):
    def fit(self):
        # step1: check inputs
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            df = pd.read_excel(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ err.message
            raise TypeError(msg)

        # step5: process
        print( '(rows, columns): ', df.shape)
        # if 'header' in self.parameters and self.parameters['header'] is not None:
        #     print 'header: ', df.columns

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs


# Prepare
class concat(BASE):
    def fit(self):
        # step1: check inputs
        objs = []
        dfs = ['df1','df2','df3']
        for f in dfs:
            df = self.inputs[f].value
            if isinstance(df,pd.DataFrame):
                objs.append(df)
        if len(objs) == 0:
            msg = "@Task #%i(%s): All the input dataframes are None objects" % (self.iblock + 1, self.Task)
            raise IOError(msg)

        # step2: assign inputs to parameters if necessary (param = @token)
        # self.paramFROMinput()
        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            df_out = pd.concat(objs,**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ err.message
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df_out)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs


# Search
class corr(BASE):
    def fit(self):
        # step1: check inputs
        self.required('df', True)
        df = self.inputs['df'].value
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            df_out = df.corr(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ err.message
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df_out)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs

# Visualize
class plot(BASE):
    def fit(self):
        # step1: check inputs
        self.required('df', True)
        df = self.inputs['df'].value
        # step2: assign inputs to parameters if necessary (param = @token)
        # self.paramFROMinput()
        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            ax = df.plot(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'fig':
                self.set_value(token, ax)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs


# Store
class to_csv(BASE):
    def fit(self):
        # step1: check inputs
        self.required('df', True)
        df = self.inputs['df'].value
        # step2: assign inputs to parameters if necessary (param = @token)
        # self.paramFROMinput()
        if 'path_or_buf' in self.parameters:
            if isinstance(self.parameters['path_or_buf'], str):
                self.parameters['path_or_buf'] = os.path.join(self.Base.output_directory,self.parameters['path_or_buf'])
        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            df.to_csv(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ err.message
            raise TypeError(msg)

        # step5: process
        print( '(rows, columns): ', df.shape)
        # if 'header' in self.parameters and self.parameters['header'] is not None:
        #     print 'headers: ', list(df.columns)

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]


        # step7: delete all inputs from memory
        del self.inputs



from __future__ import print_function
from builtins import range

import numpy as np

_LIB_NAME = "lib_jcode"
_LIB_VERSION = "v1.18.0"
_REVISION_DATE = "2012-11-09"
_AUTHOR = "Johannes Hachmann (jh@chemistry.harvard.edu)"
_DESCRIPTION = "This is the a library for general purpose functions."

# Version history timeline:
# v0.001 (2007-??-??) 
# v0.002 (2010-02-16): added tot_exec_time_str 
# v0.100 (2010-07-25,08-06): complete overhaul 
# v0.101 (2010-08-10): added intermed_exec_timing 
# v0.200 (2010-08-16): added std_datetime_str 
# v0.300 (2010-08-31): added chk_rmdir 
# v1.4.0 (2010-10-25): added wc_all; changed version format
# v1.5.0 (2010-10-31): added mksubdir_struct; modified chk_mkdir
# v1.6.0 (2010-11-01): added bin_file_format_change
# v1.7.0 (2011-04-04): added isFloat
# v1.7.1 (2011-04-04): modified isFloat, such that NaN, Nan, nan are excluded 
# v1.8.0 (2011-07-06): added target_dir_struct 
# v1.9.0 (2011-09-07): added line_count 
# v1.10.0 (2011-10-07): improve print_invoked_opts
# v1.10.1 (2011-10-13): improve formatting of print_invoked_opts
# v1.11.0 (2011-11-30): improve formatting of print_invoked_opts
# v1.12.0 (2012-04-10): add md5checksum
# v1.13.0 (2012-06-20): add filelinecount
# v1.14.0 (2012-06-20): add revdict_lookup
# v1.15.0 (2012-07-25): add intermed_process_timing
# v1.15.1 (2012-08-21): modify linecount for empty files, intermed_process_timing for zero count
# v1.15.2 (2012-10-03): modify idFloat by excluding infinity via '#IND'
# v1.16.0 (2012-10-22): add queryset_iterator
# v1.17.0 (2012-11-07,08): extend queryset_iterator with id_only and values
# v1.18.0 (2012-11-09): added list_chunks

###################################################################################################
# TASKS OF THIS LIBRARY:
# -provides general purpose functions
###################################################################################################

###################################################################################################
# TODO:
# -print lists/tuples in formatted fashion
# -review all the older routines
###################################################################################################

import sys
import os
import struct
import time
import datetime
import subprocess
import hashlib
import mmap
from numpy import fromstring

###################################################################################################

def banner(logfile, SCRIPT_NAME, SCRIPT_VERSION, REVISION_DATE, AUTHOR, DESCRIPTION,):
    """(banner):
        Banner for this little script.
    """
    str = []
    str.append("============================================================================== ")
    str.append(SCRIPT_NAME + " " + SCRIPT_VERSION + " (" + REVISION_DATE + ")")
    str.append(AUTHOR)
    str.append("============================================================================== ")
    str.append(time.ctime())
    str.append("")    
    str.append(DESCRIPTION)
    str.append("")

    print ()
    for line in str:
        print (line)
        logfile.write(line + '\n')

##################################################################################################

def print_invoked_opts(logfile,opts,commline_list=[]):
    """(print_invoked_opts):
        Prints the invoked options to stdout and the logfile.
    """    
    if len(commline_list) != 0:
        tmp_str = "Invoked command line: "
        print (tmp_str)
        logfile.write(tmp_str + '\n')
        tmp_str = ' '.join(commline_list)
        print (tmp_str)
        print 
        logfile.write(tmp_str + '\n\n')
        
    tmp_str = "Invoked options: "
    print (tmp_str)
    logfile.write(tmp_str + '\n')
    for key, value in opts.__dict__.items():
        tmp_str = '   ' + key + ': ' + str(value)   
        print (tmp_str)
        logfile.write(tmp_str + '\n')
    print()
    logfile.write('\n')

##################################################################################################

def wc_dir(dir):
    """(wc_dir):
        Returns the number of dirs in a given dir via ls -1d | wc -l. 
        Not that this becomes a rather expensive function call when dir contains many subdirs.
    """    
#TODO: take care of error for empty dirs
    tmp_str = "ls -1d " + dir + "/*/ | wc -l"
    # this is a quite new python feature and may is only available in 2.6 or so 
    # n = subprocess.getoutput(tmp_str)
    # ... and for older python versions     
    return int(subprocess.Popen(tmp_str,shell=True,stdout=subprocess.PIPE).stdout.read())

##################################################################################################

def wc_all(dir):
    """(wc_all):
        Returns the number of files and dirs in a given dir via ls -1 | wc -l. 
        Not that this becomes a rather expensive function call when dir contains many entries.
    """    
#TODO: take care of error for empty dirs
    tmp_str = "ls -1 " + dir + " | wc -l"
    # this is a quite new python feature and may is only available in 2.6 or so 
    # n = subprocess.getoutput(tmp_str)
    # ... and for older python versions     
    return int(subprocess.Popen(tmp_str,shell=True,stdout=subprocess.PIPE).stdout.read())

##################################################################################################

def line_count(file_namestr):
    """(line_count):
        Returns the number of lines in a file.
    """    
    if os.path.getsize(file_namestr) == 0:
        return 0
    with open(file_namestr) as file:
        for i, l in enumerate(file):
            pass
    return i + 1

##################################################################################################

def mksubdir_struct(dir,max_n_entries=10000,run_always=False):
    """(mksubdir_struct):
        This function takes the content of a dir and makes numbered substructure dirs with each n_entries of the original dir.
        The motivation was to have a function with limits the number of entries in a directory to a certain threshold
        (e.g., 10,000 or 30,000) in order to avoid performance issues with the OS/filesystem. 
    """
    entry_list = []
    for entry in os.listdir(dir):
        entry_list.append(entry)
    entry_list.sort()
    
    n_entries = len(entry_list)
    
    if n_entries >= max_n_entries or run_always:
        subdir_counter = 0
        subdir_entry_counter = 0
        subdir_pathstr = dir + "/%05d"  %(subdir_counter)
        
        if chk_mkdir(subdir_pathstr,True) == False:
            sys.exit("Naming conflict!")
        
        for entry in entry_list:
            tmp_str = "mv " + entry + " " + subdir_pathstr + "/." 
            os.system(tmp_str)
            subdir_entry_counter +=1
            if subdir_entry_counter >= max_n_entries:
                subdir_counter += 1
                subdir_entry_counter = 0
                subdir_pathstr = dir + "/%05d"  %(subdir_counter)
                if chk_mkdir(subdir_pathstr,True) == False:
                    sys.exit("Naming conflict!")
                
##################################################################################################

def chk_mkdir(dir,warning=False):
    """(chk_mkdir):
        This function checks whether a directory exists and if not creates it.
    """
    if not os.path.isdir(dir):
        tmp_str = "mkdir -p " + dir
        os.system(tmp_str)
    elif warning:
        return False

##################################################################################################

def chk_rmdir(dir,check='any'):
    """(chk_rmdir):
        This function checks whether a directory exists and removes it, if it is empty.
    """
    if os.path.isdir(dir):
        n_dirs = 0
        n_files = 0
        for i in os.listdir(dir):
            if os.path.isdir(dir + '/' + i):
                n_dirs += 1
            elif os.path.isfile(dir + '/' + i):
                n_files += 1
        if n_dirs == 0 and n_files == 0:
            tmp_str = "rm -rf " + dir
        elif n_dirs == 0 and check=='dirs':
            tmp_str = "rm -rf " + dir
        elif n_files == 0 and check=='files':
            tmp_str = "rm -rf " + dir
        else:
            tmp_str = " "
        os.system(tmp_str)

##################################################################################################

def chk_rmfile(file_namestr):
    """(chk_rmfile):
        This function checks whether a file is empty and if yes deletes it.
    """
    file = open(file_namestr,'r')
    test_str = file.read()
    file.close()
    if len(test_str) == 0:
        os.remove(file_namestr)
    
##################################################################################################

def target_dir_struct(target_dir_path, maxitems = 10000, digits=5):
    """(target_dir_struct):
        This function checks whether a target dir exists and establishes/checks the subdir structure.
    """
    # check if target_dir exists and if not create it
    chk_mkdir(target_dir_path)
    # establish target_dir structure
    # 1) get all the present subdirs
    target_subdir_list = [] # fill with all present subfolders
    for i in os.listdir(target_dir_path):
        if os.path.isdir(target_dir_path + '/' + i) and i not in target_subdir_list:
            target_subdir_list.append(i)
    # 2a) if there are no subfolders present
    if len(target_subdir_list)==0:
        target_subdir = 0   # this is the highest folder
        target_subdir_n = 0 # this is the number of items in it
    # 2b) if there are subfolders present    
    else:
        target_subdir_list.sort()
        target_subdir = int(target_subdir_list[-1]) # pick the highest folder
        target_subdir_n = wc_all(target_dir_path + '/' + target_subdir_list[-1])
        if target_subdir_n >= maxitems:     # this limit is more important for folders rather than files (in this case tarballs); but we do it anyways
            target_subdir += 1
            target_subdir_n = 0

    target_subdir_pathstr = target_dir_path + '/' + '{num:{fill}{width}}'.format(num=target_subdir, fill='0', width=digits)
#    target_subdir_pathstr = target_dir_path + "/%05d"  %(target_subdir)
    chk_mkdir(target_subdir_pathstr)
    return target_subdir, target_subdir_n, target_subdir_pathstr
    
##################################################################################################

def mv2subdir_struct(source_dir_pathstr, target_subdir, target_subdir_n, target_subdir_pathstr, maxitems = 10000):
    """(mv2subdir_struct):
        This function moves a source folder into a target subdir structure and updates it.
    """
    # move
    tmp_str = 'mv ' + source_dir_pathstr + ' ' + target_subdir_pathstr + '/. ' 
    os.system(tmp_str)
    target_subdir_n += 1

    # check if limit is reached
    if target_subdir_n >= maxitems:     # this limit is more important for folders rather than files (in this case tarballs); but we do it anyways
        target_subdir += 1
        target_subdir_n = 0
        
        # make new target subdir
        tmp_str = target_subdir_pathstr.split('/')[-1]
        digits = len(tmp_str)
        target_subdir_pathstr = target_subdir_pathstr[:-digits] + '{num:{fill}{width}}'.format(num=target_subdir, fill='0', width=digits)
        chk_mkdir(target_subdir_pathstr)
    return target_subdir, target_subdir_n, target_subdir_pathstr
    
##################################################################################################

def std_datetime_str(mode='datetime'):
    """(std_time_str):
        This function gives out the formatted time as a standard string, i.e., YYYY-MM-DD hh:mm:ss.
    """
    if mode == 'datetime':
        return str(datetime.datetime.now())[:19]
    elif mode == 'date':
        return str(datetime.datetime.now())[:10]
    elif mode == 'time':
        return str(datetime.datetime.now())[11:19]
    elif mode == 'datetime_ms':
        return str(datetime.datetime.now())
    elif mode == 'time_ms':
        return str(datetime.datetime.now())[11:]
    else:
        sys.exit("Invalid mode!")

##################################################################################################
def tot_exec_time_str(time_start):
    """(tot_exec_time_str):
        This function gives out the formatted time string.
    """
    time_end = time.time()
    exec_time = time_end-time_start
    tmp_str = "execution time: %0.2fs (%dh %dm %0.2fs)" %(exec_time, exec_time/3600, (exec_time%3600)/60,(exec_time%3600)%60)
    return tmp_str

##################################################################################################

def intermed_exec_timing(time_start,intermed_n,total_n,n_str="n"):
    """(intermed_exec_timing):
        This function gives out the intermediate timing, speed, pace, projected remaining and end time.
    """
    tmp_time = time.time()
    tmp_exec_time = tmp_time-time_start
    sec_per_n = 1.0*tmp_exec_time/intermed_n
    n_per_hour = 3600.0/sec_per_n
    proj_rest_sec = sec_per_n*(total_n-intermed_n)
    proj_end_time = int(round(tmp_time + proj_rest_sec))
    tmp_str = "   Current speed: %0.2f " %(n_per_hour)
    tmp_str += n_str + "'s/hour; current pace: %0.3f " %(sec_per_n)
    tmp_str += "sec/" + n_str + "\n" 
#    tmp_str +="   Projected remaining time: %0.2fs (%dh %dm %0.2fs) " %(proj_rest_sec, proj_rest_sec/3600, (proj_rest_sec%3600)/60,(proj_rest_sec%3600)%60)
    tmp_str +="   Projected remaining time: %0.2fs (%dh %dm %0.2fs) \n" %(proj_rest_sec, proj_rest_sec/3600, (proj_rest_sec%3600)/60,(proj_rest_sec%3600)%60)
    tmp_str +="   Projected end time: " + time.ctime(proj_end_time) 
    return tmp_str

##################################################################################################

def intermed_process_timing(time_start,process_n,intermed_n,total_n,n_str="n"):
    """(intermed_process_timing):
        This function gives out the intermediate timing, speed, pace, projected remaining and end time of a particular process with restarted time.
    """
    tmp_time = time.time()
    tmp_exec_time = tmp_time-time_start
    if process_n == 0:
        return ''
    
    sec_per_n = 1.0*tmp_exec_time/process_n
    n_per_hour = 3600.0/sec_per_n
    proj_rest_sec = sec_per_n*(total_n-intermed_n)
    proj_end_time = int(round(tmp_time + proj_rest_sec))
    tmp_str = "   Current speed: %0.2f " %(n_per_hour)
    tmp_str += n_str + "'s/hour; current pace: %0.3f " %(sec_per_n)
    tmp_str += "sec/" + n_str + "\n" 
#    tmp_str +="   Projected remaining time: %0.2fs (%dh %dm %0.2fs) " %(proj_rest_sec, proj_rest_sec/3600, (proj_rest_sec%3600)/60,(proj_rest_sec%3600)%60)
    tmp_str +="   Projected remaining time: %0.2fs (%dh %dm %0.2fs) \n" %(proj_rest_sec, proj_rest_sec/3600, (proj_rest_sec%3600)/60,(proj_rest_sec%3600)%60)
    tmp_str +="   Projected end time: " + time.ctime(proj_end_time) 
    return tmp_str

##################################################################################################

def timeit(func):
    """(timeit):
        Annotate a function with its elapsed execution time.
    """
    def timed_func(*args, **kwargs):
        t1 = time.time()
        
        try:
            func(*args, **kwargs)
        finally:
            t2 = time.time()

        timed_func.func_time = ((t2 - t1) / 60.0, t2 - t1)

        if __debug__:
            sys.stdout.write("%s took %0.3fm %0.3fs %0.3fms\n" % (
                func.func_name,
                timed_func.func_time[0],
                timed_func.func_time[1],
            ))

    return timed_func

###################################################################################################

def dsu_sort(list, index, reverse=False):
    """(dsu_sort):
    """
# TODO: infoline
    for i, e in enumerate(list):
        list[i] = (e[index], e)
    if reverse:
        list.sort(reverse=True)
    else:
        list.sort()
    for i, e in enumerate(list):
        list[i] = e[1]
    return list

###################################################################################################

def dsu_sort2(list, index, reverse=False):
    """(dsu_sort2):
        This function sorts only based on the primary element, not on secondary elements in case of equality.
    """
    for i, e in enumerate(list):
        list[i] = e[index]
    if reverse:
        list.sort(reverse=True)
    else:
        list.sort()
    for i, e in enumerate(list):
        list[i] = e[1]
    return list

###################################################################################################

def bin_file_format_change(infile_namestr,outfile_namestr,mode):
    """(bin_file_format_change):
        This function reads in a binary file of a certain format, converts it, and gives out a binary file of the new format.
    """
# TODO: open file and read in
# TODO: this needs to be a binary read
    infile = open(infile_namestr,'rb',0)
    in_bin_str = infile.read()
    infile.close()
# TODO: change binary format
# TODO: this needs to be a binary write
    outfile = open(outfile_namestr,'wb',0)    
    if mode == 'sp2dp':
        in_bin = fromstring(in_bin_str,np.float32)
        sys.exit()
    elif mode == 'dp2sp':
        in_bin = fromstring(in_bin_str,np.float64)
        sys.exit()
    else:
        sys.exit("Unknown binary format conversion mode.")
# TODO: make new file and dump
    
    outfile.close()

###################################################################################################

def isFloat(x):
    if x in ('NAN','NaN','Nan','nan'):
        return 0
    elif '#IND' in x:
        return 0
    try:
        float(x)
        return 1
    except:
        return 0

###################################################################################################

def md5checksum(file_path, blocksize=8192): # 4kB blocks
    """(md5checksum):
        Compute md5 hash of the specified file.
    """
    file = open(file_path, 'rb')
    md5sum = hashlib.md5()
    while True:
        data = file.read(blocksize)
        if not data:
            break
        md5sum.update(data)
    file.close()
    return md5sum.hexdigest()

###################################################################################################

def filelinecount(filename):
    """(filelinecount):
        Counts the number of lines in a file.
    """
    f = open(filename, "r+")
    buf = mmap.mmap(f.fileno(), 0)
    lines = 0
    readline = buf.readline
    while readline():
        lines += 1
    return lines

###################################################################################################

def revdict_lookup(dict,lookup_val):
    """(revdict_lookup):
        Performs a reverse dictionary lookup. Careful: only returns first match, but there may be others.
    """    
    key = (key for key,value in dict.items() if value==lookup_val).next()
    return key


###################################################################################################

def queryset_iterator(queryset, chunksize=1000, reverse=False, id_only=False, values= False):
    """(queryset_iterator):
        Django incremental queryset iterator.
        Found on: http://www.poeschko.com/2012/02/memory-efficient-django-queries/
    """    
    ordering = '-' if reverse else ''
    queryset = queryset.order_by(ordering + 'pk')
    last_pk = None
    new_items = True
    while new_items:
        new_items = False
        chunk = queryset
        if last_pk is not None:
            func = 'lt' if reverse else 'gt'
            chunk = chunk.filter(**{'pk__' + func: last_pk})
        chunk = chunk[:chunksize]
        if id_only:
            chunk = chunk.values('pk')            
        row = None
        for row in chunk:
            yield row
        if row is not None:
            if id_only or values:
                last_pk = row['pk']                
            else:
                last_pk = row.pk
            new_items = True

###################################################################################################

def list_chunks(l, n):
    """ Yield successive n-sized chunks from l.
        Found on: http://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks-in-python
    """
    for i in range(0, len(l), n):
        yield l[i:i+n]


import pandas as pd
import numpy as np
import warnings
import inspect

from ...utils.utilities import list_del_indices

class Preprocessor(object):
    def imputer(self, transformer, df, method):
        """ keep track of features (columns) that can be removed or changed in the
            Imputer by transforming data back to pandas dataframe structure. This happens based on
            the "statistics_" attribute of Imputer.

            this function also distinguish different methods of Imputer.
        Parameters
        ----------
        transformer: sklearn Imputer class
             The class with adjusted parameters.

        df: Pandas dataframe
            The dataframe that imputer is going to deal with.

        method: string
            sklearn methods for imputer: 'fit_transform' and 'transform'

        Returns
        -------
        transformed data frame
        fitted imputer class
        """

        df_columns = list(df.columns)
        if method == 'fit_transform':
            df = transformer.fit_transform(df)
        elif method == 'transform':
            df = transformer.transform(df)
        else:
            msg = "@Task #%i(%s): The passed method is not valid. It can be any of fit_transform, transform." % (self.iblock, self.Task)
            raise NameError(msg)
        if df.shape[1] == 0:
            warnings.warn("@Task #%i(%s): empty dataframe: all columns have been removed" \
                          %(self.iblock + 1, self.Task), Warning)
        else:
            stats = transformer.statistics_
            nan_ind = [i for i, val in enumerate(stats) if np.isnan(val)]
            df_columns = list_del_indices(df_columns, nan_ind)
            df = pd.DataFrame(df, columns=df_columns)
        return transformer, df

    def fit_transform_inverse(self, api, df, method, header=True):
        """ keep track of features (columns) that can be removed or changed in the
            preprocessor by transforming data back to pandas dataframe.

            this function also distinguish different methods of APIs.

        Parameters
        ----------
        api: an instance of sklearn class
            The class with adjusted parameters.

        df: Pandas dataframe
            The dataframe that api is going to deal with

        method: string
            sklearn methods for the api: 'fit_transform', 'transform', 'inverse_transform'

        Returns
        -------
        transformed data frame
        fitted api class

        """

        df_columns = list(df.columns)

        if method == 'fit_transform':
            df = api.fit_transform(df)
        elif method == 'transform':
            df = api.transform(df)
        elif method == 'inverse_transform':
            df = api.inverse_transform(df)
        if header:
            if df.shape[1] == 0:
                warnings.warn("@Task #%i(%s): empty dataframe - all columns have been removed"
                              % (self.iblock + 1, self.Task), Warning)
            if df.shape[1] == len(df_columns):
                df = pd.DataFrame(df, columns=df_columns)
            else:
                df = pd.DataFrame(df)
                warnings.warn("@Task #%i(%s): headers are untrackable - number of columns before and after transform doesn't match" \
                    %(self.iblock + 1, self.Task), Warning)
        return api, df

    #unused
    def selector_dataframe(self, transformer, df, tf):
        """ keep track of features (columns) that can be removed or changed in the
            VarianceThreshold by transforming data back to pandas dataframe structure.
            This happens based on the "get_support" method of selector.

        Parameters
        ----------
        transformer: sklearn VarianceThreshold class
             The class with adjusted parameters.

        df: Pandas dataframe
            data frame

        tf: Pandas dataframe
            target frame

        Returns
        -------
        transformed data frame
        fitted imputer class
        """
        df_columns = list(df.columns)
        df = transformer.fit_transform(df, tf)
        if df.shape[1] == 0:
            warnings.warn("empty dataframe: all columns have been removed", Warning)
            return transformer, df
        else:
            retained_features_ind = transformer.get_support(True)
            df_columns = [df_columns[i] for i in retained_features_ind]
            df = pd.DataFrame(df, columns=df_columns)
            return df

class Regressor(object):
    def sklearn_training(self,model,X,y):
        model.fit(X, y)
        r2score_training = model.score(X, y)
        return model, r2score_training

    def train(self, model, X, Y):
        if model[1][2] == 'sklearn':
            model[0].fit(X, Y)
            model_trained = model[0]
            r2score_training = model[0].score(X, Y)
        elif model[1][2] == 'cheml' and model[1][3] == 'NN_PSGD':
            from cheml.nn import nn_psgd
            ev_inds = np.random.choice(np.arange(X.shape[0]),int(0.2*X.shape[0])+1,replace=False)
            tr_inds = np.delete(np.arange(X.shape[0]),ev_inds)
            x_tr = X.loc[tr_inds,:]
            x_ev = X.loc[ev_inds,:]
            y_tr = Y.loc[tr_inds,:]
            y_ev = Y.loc[ev_inds,:]
            model_trained = nn_psgd.train(x_tr, x_ev, y_tr, y_ev, **model[0])
            from sklearn.metrics import r2_score
            r2score_training = r2_score(y_true=Y, y_pred=nn_psgd.output(X,model_trained), multioutput='raw_values')
        return model_trained,r2score_training


    def predict(self, model, X):
        if model[1][2] == 'sklearn':
            Y_pred = pd.DataFrame(model[0].predict(X))
        elif model[1][2] == 'cheml' and model[1][3] == 'NN_PSGD':
            from cheml.nn import nn_psgd
            Y_pred = nn_psgd.output(X.values, model[0])
        return Y_pred

class Evaluator(object):
    def _reg_evaluation_params(self):
        self.params = {}
        self.evaluator = {}
        if 'r2_score' in self.parameters and self.parameters['r2_score'] == True:
            metric = 'r2_score'
            self.params[metric] = {}
            if 'r2_sample_weight' in self.parameters:
                self.params[metric]['sample_weight'] = self.parameters['r2_sample_weight']
            if 'r2_multioutput' in self.parameters:
                self.params[metric]['multioutput'] = self.parameters['r2_multioutput']

            # import module and make APIs
            from sklearn.metrics import r2_score
            self.evaluator[metric] = lambda Y, Y_pred: r2_score(y_true=Y, y_pred=Y_pred, **self.params[metric])

        if 'mean_absolute_error' in self.parameters and self.parameters['mean_absolute_error'] == True:
            metric = 'mean_absolute_error'
            self.params[metric] = {}
            if 'mae_sample_weight' in self.parameters:
                self.params[metric]['sample_weight'] = self.parameters['mae_sample_weight']
            if 'mae_multioutput' in self.parameters:
                self.params[metric]['multioutput'] = self.parameters['mae_multioutput']

            # import module and make APIs
            from sklearn.metrics import mean_absolute_error
            self.evaluator[metric] = lambda Y, Y_pred: mean_absolute_error(y_true=Y, y_pred=Y_pred, **self.params[metric])

        if 'median_absolute_error' in self.parameters and self.parameters['median_absolute_error'] == True:
            metric = 'median_absolute_error'
            self.params[metric] = {}

            # import module and make APIs
            from sklearn.metrics import median_absolute_error
            self.evaluator[metric] = lambda Y, Y_pred: median_absolute_error(y_true=Y, y_pred=Y_pred)#, **self.params[metric])

        if 'mean_squared_error' in self.parameters and self.parameters['mean_squared_error'] == True:
            metric = 'mean_squared_error'
            self.params[metric] = {}
            if 'mse_sample_weight' in self.parameters:
                self.params[metric]['sample_weight'] = self.parameters['mse_sample_weight']
            if 'mse_multioutput' in self.parameters:
                self.params[metric]['multioutput'] = self.parameters['mse_multioutput']

            # import module and make APIs
            from sklearn.metrics import mean_squared_error
            self.evaluator[metric] = lambda Y, Y_pred: mean_squared_error(y_true=Y, y_pred=Y_pred, **self.params[metric])

        if 'root_mean_squared_error' in self.parameters and self.parameters['root_mean_squared_error'] == True:
            metric = 'root_mean_squared_error'
            self.params[metric] = {}
            if 'rmse_sample_weight' in self.parameters:
                self.params[metric]['sample_weight'] = self.parameters['rmse_sample_weight']
            if 'rmse_multioutput' in self.parameters:
                self.params[metric]['multioutput'] = self.parameters['rmse_multioutput']

            # import module and make APIs
            from sklearn.metrics import mean_squared_error
            self.evaluator[metric] = lambda Y, Y_pred: np.sqrt(mean_squared_error(y_true=Y, y_pred=Y_pred, **self.params[metric]))

        if 'explained_variance_score' in self.parameters and self.parameters['explained_variance_score'] == True:
            metric = 'explained_variance_score'
            self.params[metric] = {}
            if 'ev_sample_weight' in self.parameters:
                self.params[metric]['sample_weight'] = self.parameters['ev_sample_weight']
            if 'ev_multioutput' in self.parameters:
                self.params[metric]['multioutput'] = self.parameters['ev_multioutput']

            # import module and make APIs
            from sklearn.metrics import explained_variance_score
            self.evaluator[metric] = lambda Y, Y_pred: explained_variance_score(y_true=Y, y_pred=Y_pred, **self.params[metric])

    def _reg_evaluate(self, Y, Y_pred, evaluator):
        if Y.shape[1]>1:
            msg = "@Task #%i(%s): be aware of 'multioutput' behavior of selected metrics" \
                          %(self.iblock + 1, self.Task)
            warnings.warn(msg, Warning)
        self.results = {}
        for metric in evaluator:
            try:
                self.results[metric] = [evaluator[metric](Y, Y_pred)]
            except Exception as err:
                msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                    err).__name__ + ': ' + err.message
                raise TypeError(msg)


def Fit(fn):
    """ Fit with a wrapper """
    def wrapper(self):
        self.paramFROMinput()
        if 'track_header' in self.parameters:
            self.header = self.parameters.pop('track_header')
        else:
            self.header = True
        if 'func_method' in self.parameters:
            self.method = self.parameters.pop('func_method')
        else:
            self.method = None
        available_methods = self.metadata.WParameters.func_method.options
        if self.method not in available_methods:
            msg = "@Task #%i(%s): The method '%s' is not available for the function '%s'." % (
                self.iblock, self.Task,self.method,self.Function)
            raise NameError(msg)
        else:
            if self.method is None:
                api = self.import_sklearn()
                self.set_value('api', api)
                fn(self)
            elif self.method is 'fit_transform':
                api = self.import_sklearn()
                self.required('df', req=True)
                df = self.inputs['df'].value
                df = api.fit_transform(df)
                self.set_value('api', api)
                self.set_value('df', df)
                fn(self)
            elif self.method is 'transform':
                self.required('df', req=True)
                self.required('api', req=True)
                df = self.inputs['df'].value
                api = self.inputs['api'].value
                df = api.transform(df)
                self.set_value('api', api)
                self.set_value('df', df)
                fn(self)
            elif self.method is 'inverse_transform':
                self.required('df', req=True)
                self.required('api', req=True)
                df = self.inputs['df'].value
                api = self.inputs['api'].value
                df = api.inverse_transform(df)
                self.set_value('api', api)
                self.set_value('df', df)
            elif self.method is 'fit':
                api = self.import_sklearn()
                self.required('dfx', req=True)
                self.required('dfy', req=True)
                dfx = self.inputs['dfx'].value
                dfy = self.inputs['dfy'].value
                api.fit(dfx,dfy)
                dfy_predict = api.predict(dfx)
                self.set_value('api', api)
                self.set_value('dfy_predict', dfy_predict)
                fn(self)
            elif self.method is 'predict':
                self.required('dfx', req=True)
                self.required('api', req=True)
                dfx = self.inputs['dfx'].value
                api = self.inputs['api'].value
                dfy_predict = api.predict(dfx)
                self.set_value('api', api)
                self.set_value('dfy_predict', dfy_predict)
                fn(self)

        self.Send()

        # delete all inputs
        del self.inputs
